# -*- coding: utf-8 -*-

from typing import Union
from contextlib import contextmanager

from selenium.webdriver import (
    Remote as _Remote,
    Chrome as _Chrome,
    Firefox as _Firefox,
    PhantomJS as _PhantomJS,
    Opera as _Opera,
    Safari as _Safari,
)

from pagium import utils


class WEbDriverPollingMixin:

    def __init__(self, *args, **kwargs):
        self._polling_timeout = kwargs.pop('polling_timeout', None)
        self._polling_delay = kwargs.pop('polling_delay', utils.DEFAULT_POLLING_DELAY)
        self._enable_polling = True if self._polling_timeout else False

        self._implicitly_wait = 0
        self._set_script_timeout = 0

        with self.disable_polling():
            super(WEbDriverPollingMixin, self).__init__(*args, **kwargs)

    @property
    def polling_timeout(self):
        return self._polling_timeout

    @property
    def polling_delay(self):
        return self._polling_delay

    @contextmanager
    def disable_polling(self, *, force=False):
        implicitly_wait = script_timeout = 0

        if force:
            implicitly_wait, script_timeout = self._implicitly_wait, self._set_script_timeout
            self.implicitly_wait(0)
            self.set_script_timeout(0)

        ep = self._enable_polling
        self._enable_polling = False

        try:
            yield
        finally:
            self._enable_polling = ep

            if force:
                self.implicitly_wait(implicitly_wait)
                self.set_script_timeout(script_timeout)

    @contextmanager
    def enable_polling(self,
                       timeout: Union[int, float] = utils.DEFAULT_POLLING_TIMEOUT,
                       delay: Union[int, float] = utils.DEFAULT_POLLING_DELAY):
        pt, pd, ep = self._polling_timeout, self._polling_delay, self._enable_polling
        self._polling_timeout, self._polling_delay, self._enable_polling = timeout, delay, True

        try:
            yield
        finally:
            self._polling_timeout, self._polling_delay, self._enable_polling = pt, pd, ep

    def execute(self, *args, **kwargs):
        if self._enable_polling:
            execute = utils.polling(
                super(WEbDriverPollingMixin, self).execute,
                timeout=self._polling_timeout, delay=self._polling_delay,
            )
        else:
            execute = super(WEbDriverPollingMixin, self).execute

        return execute(*args, **kwargs)

    def implicitly_wait(self, wait_timeout):
        self._implicitly_wait = wait_timeout
        super(WEbDriverPollingMixin, self).implicitly_wait(self._implicitly_wait)

    def set_script_timeout(self, wait_timeout):
        self._set_script_timeout = wait_timeout
        super(WEbDriverPollingMixin, self).set_script_timeout(self._set_script_timeout)


class Remote(WEbDriverPollingMixin, _Remote):
    """
    >>> wd = Remote(
    ... command_executor='http://localhost:4444/wd/hub',
    ... desired_capabilities={'browserName': 'chrome'},
    ... )

    >>> wd.quit()
    """
    pass


class Chrome(WEbDriverPollingMixin, _Chrome):
    pass


class Firefox(WEbDriverPollingMixin, _Firefox):
    pass


class PhantomJS(WEbDriverPollingMixin, _PhantomJS):
    pass


class Opera(WEbDriverPollingMixin, _Opera):
    pass


class Safari(WEbDriverPollingMixin, _Safari):
    pass





"""
>>> from pagium.webdriver import Remote
>>> from selenium.webdriver.common.by import By
>>> from selenium.webdriver.common.keys import Keys

>>> wd = Remote(
... command_executor='http://localhost:4444/wd/hub',
... desired_capabilities={'browserName': 'chrome'},
... polling_timeout=10,
... polling_delay=0.5,
... )

>>> class GooglePageTestBasic(Page):
...     search = PageElement(by=By.NAME, value='q')

>>> with GooglePageTestBasic(wd, 'https://google.com') as page:
...     page.search.send_keys(*'python selenium' + Keys.ENTER)

>>> with GooglePageTestBasic(wd, 'https://google.com') as page:
...     assert page.search.exists() == True

>>> class SearchInput(WebElement):
...     def fill(self, text):
...         self.send_keys(*text + Keys.ENTER)

>>> class GooglePageTestControl(Page):
...     search = PageElement(SearchInput, by=By.NAME, value='q')

>>> with GooglePageTestControl(wd, 'https://google.com') as page:
...     page.search.fill('python selenium')

>>> class SearchForm(WebElement):
...     input = PageElement(SearchInput, by=By.NAME, value='q')

>>> class GooglePageTestSeparatedComponent(Page):
...     search_form = PageElement(SearchForm)

>>> with GooglePageTestSeparatedComponent(wd, 'https://google.com') as page:
...     page.search_form.input.fill('python selenium')

>>> wd.quit()
"""

from typing import Union, Callable
from urllib.parse import urljoin

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement


class Page:

    __path__ = None

    def __init__(self, parent: Union[WebDriver, WebElement], url: str, **options):
        """
        >>> assert Page(object, 'http://google.com').url ==  'http://google.com'

        >>> class TestPage(Page):
        ...     __path__ = '/test/{param}'

        >>> assert TestPage(
        ... object, 'http://google.com', path={'param': 'one'},
        ... ).url == 'http://google.com/test/one'

        >>> assert TestPage(object, '').parent == object

        >>> assert 'test' in TestPage(object, '', test=1).options
        """
        self._url = url

        self._parent = parent
        self._options = options

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args, **kwargs):
        self.close()

    def __getattr__(self, item):
        return getattr(self._parent, item)

    def __repr__(self):
        return f'<{self.__class__.__name__} on {self.url}>'

    @property
    def parent(self):
        return self._parent

    @property
    def options(self):
        return self._options

    @property
    def text(self) -> str:
        if isinstance(self._parent, WebDriver):
            return str(self._parent.find_element_by_tag_name('body').text)
        return str(self._parent.text)

    @property
    def url(self):
        if self.__path__ is not None:
            return urljoin(
                self._url, self.__path__.format(
                    **self._options.get('path', {}),
                ),
            )

        return self._url

    def open(self):
        if isinstance(self._parent, WebDriver):
            self._parent.get(self.url)
        else:
            raise AssertionError(
                'Can not open page because parent is not instance of WebDriver object',
            )

    def close(self):
        pass


class PageElement:

    def __init__(self,
                 we_class: type = None,
                 by: str = None,
                 value: str = None,
                 is_list: bool = False,
                 hook: Callable = None):
        if we_class is not None:
            if not issubclass(we_class, WebElement):
                raise AssertionError(
                    'Web element type can be subclass of WebElement only',
                )

        if by is None and value is None and we_class is None:
            raise AssertionError(
                '"by", "value" or "we_type" is required params for search web element',
            )

        self._we_class = we_class
        self._by = by
        self._value = value
        self._is_list = is_list
        self._hook = hook

    def __repr__(self):
        we_class = self._we_class or WebElement
        return f'{we_class.__name__}: {self._by}={self._value} '

    def __get__(self, instance: Union[Page, WebElement], owner: type):
        parent = instance

        if isinstance(instance, Page):
            parent = instance.parent

        lazy_web_element = LazyWebElement(self, parent)

        if callable(self._hook):
            return self._hook(lazy_web_element)

        return lazy_web_element

    @property
    def by(self):
        return self._by

    @property
    def we_class(self):
        return self._we_class

    @property
    def value(self):
        return self._value

    @property
    def is_list(self):
        return self._is_list

    def get(self, parent: Union[WebDriver, WebElement]) -> Union[WebElement, list]:
        if self._by is None and self._value is None:
            if self._we_class is None:
                raise AssertionError(
                    'Can not create web element container, use "we_type" param for resolve it',
                )
            container = self._we_class(parent, parent.session_id)
            container.find_element = parent.find_element
            container.find_elements = parent.find_elements

            return container

        if self._is_list:
            web_element = parent.find_elements(by=self._by, value=self._value)

            if self._we_class is not None:
                for we in web_element:
                    we.__class__ = self._we_class
        else:
            web_element = parent.find_element(by=self._by, value=self._value)

            if self._we_class is not None:
                web_element.__class__ = self._we_class

        return web_element


class LazyWebElement:

    def __init__(self, element: PageElement, parent: Union[WebDriver, WebElement]):
        self._page_element = element
        self._parent = parent
        self._web_element = None

    def __repr__(self):
        self._search()
        return f'{self.__class__.__name__} -> {repr(self._web_element)}'

    def __iter__(self):
        self._search()
        return iter(self._web_element)

    def __getattr__(self, item):
        self._search()
        return getattr(self._web_element, item)

    def __getitem__(self, item):
        self._search()
        return self._web_element.__getitem__(item)

    def __len__(self):
        self._search()
        if self._page_element.is_list:
            return self._web_element.__len__()

        return 1

    def _search(self):
        if self._web_element is None:
            self._web_element = self._page_element.get(self._parent)

    @property
    def parent(self):
        return self._parent

    def exists(self, count: int = 1) -> bool:
        self.refresh()

        try:
            self._search()
        except WebDriverException:
            return False

        if self._page_element.is_list:
            return len(self._web_element) >= count

        return self._web_element.is_displayed()

    def refresh(self):
        self._web_element = None



# -*- coding: utf-8 -*-

"""
Python page object patter realization for selenium library.
"""

from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys

from pagium.page import Page, PageElement, WebElement
from pagium.webdriver import Remote, Chrome, Firefox, PhantomJS, Opera, Safari


__all__ = [
    'By',
    'Page',
    'PageElement',
    'WebElement',
    'Remote',
    'Chrome',
    'Firefox',
    'PhantomJS',
    'Opera',
    'Safari',
    'Keys',
]



# -*- coding: utf-8 -*-

import time
import socket
from functools import wraps
from typing import Union, Callable
from http.client import HTTPException

from selenium.common.exceptions import WebDriverException
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement


DEFAULT_POLLING_EXCEPTIONS = (
    IOError,
    OSError,
    HTTPException,
    WebDriverException,
)


DEFAULT_POLLING_TIMEOUT = 30
DEFAULT_POLLING_DELAY = 0.5


def polling(callback: Callable,
            timeout: Union[float, int] = DEFAULT_POLLING_TIMEOUT,
            delay: Union[float, int] = DEFAULT_POLLING_DELAY,
            except_exceptions=DEFAULT_POLLING_EXCEPTIONS):
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            t_start = time.time()
            error = None

            while time.time() <= t_start + timeout:
                try:
                    return f(*args, **kwargs)
                except socket.error:
                    raise
                except except_exceptions as e:
                    error = e

                    if delay:
                        time.sleep(delay)

                    continue
            else:
                raise error

        return wrapped
    return wrapper(callback)


def waiting_for(callback: Callable,
                timeout: Union[float, int] = DEFAULT_POLLING_TIMEOUT,
                delay: Union[float, int] = DEFAULT_POLLING_DELAY,
                raise_exc=None,
                message: str = None,
                args: Union[list, tuple] = None,
                kwargs: dict = None):
    result = None

    args = args or tuple()
    kwargs = kwargs or dict()

    timeout = timeout or 0
    message = message or 'Timeout "{}" exceeded'.format(timeout)

    if timeout:
        t_start = time.time()

        while time.time() <= t_start + timeout:
            result = callback(*args, **kwargs)

            if result:
                return result

            if delay:
                time.sleep(delay)
        else:
            if raise_exc:
                raise raise_exc(message)

            return result

    result = callback(*args, **kwargs)

    if result:
        return result

    if raise_exc:
        raise raise_exc(message)

    return result


def get_driver(instance: Union[WebDriver, WebElement]):
    if isinstance(instance, WebDriver):
        return instance

    while not isinstance(instance, WebDriver):
        instance = instance.parent

    return instance



# -*- coding: utf-8 -*-

from selenium.webdriver.common.keys import Keys

from pagium.page import LazyWebElement


class BasePageElementHook:

    def __init__(self, web_element: LazyWebElement):
        self.web_element = web_element

    def __repr__(self):
        return f'Hook: {self.__class__.__name__}'


class ClickActionHook(BasePageElementHook):
    """
    >>> from pagium.webdriver import Remote
    >>> from pagium.page import PageElement, Page
    >>> from selenium.webdriver.common.by import By

    >>> wd = Remote(
    ... command_executor='http://localhost:4444/wd/hub',
    ... desired_capabilities={'browserName': 'chrome'},
    ... polling_timeout=10,
    ... polling_delay=0.5,
    ... )

    >>> class GooglePageTestBasic(Page):
    ...     search = PageElement(by=By.NAME, value='q')
    ...     submit = PageElement(by=By.NAME, value='btnK', hook=ClickActionHook)

    >>> with GooglePageTestBasic(wd, 'https://google.com') as page:
    ...     page.search.send_keys(*'python selenium')
    ...     page.submit()

    >>> wd.quit()
    """

    def __call__(self):
        self.web_element.click()


class SendEnterHook(BasePageElementHook):
    """
    >>> from pagium.webdriver import Remote
    >>> from pagium.page import PageElement, Page
    >>> from selenium.webdriver.common.by import By

    >>> wd = Remote(
    ... command_executor='http://localhost:4444/wd/hub',
    ... desired_capabilities={'browserName': 'chrome'},
    ... polling_timeout=10,
    ... polling_delay=0.5,
    ... )

    >>> class GooglePageTestBasic(Page):
    ...     search = PageElement(by=By.NAME, value='q', hook=SendEnterHook)

    >>> with GooglePageTestBasic(wd, 'https://google.com') as page:
    ...     page.search('python selenium')

    >>> wd.quit()
    """

    def __call__(self, text=None):
        if text is None:
            text = Keys.ENTER
        else:
            text = text + Keys.ENTER

        self.web_element.send_keys(*text)



# -*- coding: utf-8 -*-

from selenium.webdriver.remote.webelement import WebElement
from selenium.common.exceptions import NoSuchElementException


class Input(WebElement):

    @property
    def value(self):
        return self.get_attribute('value')

    @property
    def placeholder(self):
        return self.get_attribute('placeholder')

    def fill(self, text: str):
        self.clear()
        self.send_keys(*text)


class Link(WebElement):

    @property
    def href(self):
        return self.get_attribute('href')


class Image(WebElement):

    @property
    def src(self):
        return self.get_attribute('src')


class Select(WebElement):

    @property
    def options(self):
        return self.find_elements_by_tag_name('option')

    @property
    def selected(self):
        return next(iter([option for option in self.options if option.is_selected()]), None)

    def select(self, value, by='value'):
        for option in self.options:
            if (by == 'value' and option.get_attribute('value') == value) or (by == 'text' and option.text == value):
                option.click()
                break
        else:
            raise NoSuchElementException(f'Option "{value}" was not found')


class Checkbox(WebElement):

    def is_checked(self):
        return self.is_selected()

    def check(self):
        if not self.is_checked():
            self.click()

    def uncheck(self):
        if self.is_checked():
            self.click()



# -*- coding: utf8 -*-

import re
from typing import Union
from urllib.parse import urlparse

from hamcrest.core.base_matcher import BaseMatcher
from selenium.webdriver.remote.webelement import WebElement

from pagium import utils
from pagium.page import Page, LazyWebElement


DEFAULT_TIMEOUT = 30
DEFAULT_DELAY = 0.5


class _BasePagiumMatcher(BaseMatcher):

    def __init__(self, *, timeout: int = DEFAULT_TIMEOUT, delay: float = DEFAULT_DELAY):
        self.timeout = timeout
        self.delay = delay

    def __matches__(self, *args, **kwargs):
        pass

    def _matches(self, *args, **kwargs):
        return utils.waiting_for(
            self.__matches__, args=args, kwargs=kwargs,
            timeout=self.timeout, delay=self.delay,
        )

    def _create_message(self, text, **params):
        params.update(timeout=self.timeout, delay=self.delay)
        params_string = ', '.join(f'{k}={v}' for k, v in params.items())

        return f'{text} ({params_string})'


class _HasText(_BasePagiumMatcher):

    def __init__(self, text: str, **kwargs):
        super(_HasText, self).__init__(**kwargs)

        self.text = text

    def __matches__(self, instance: Union[Page, WebElement, LazyWebElement]):
        return str(self.text).lower() in str(instance.text).lower()

    def describe_to(self, description):
        description.append_text(
            self._create_message(f'Text "{self.text}" exists'),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text('was not found')


has_text = _HasText


class _ElementExists(_BasePagiumMatcher):

    def __init__(self, count: int = 1, **kwargs):
        super(_ElementExists, self).__init__(**kwargs)

        self.count = count

    def __matches__(self, lazy_web_element: LazyWebElement):
        driver = utils.get_driver(lazy_web_element.parent)

        if hasattr(driver, 'disable_polling'):
            with driver.disable_polling(force=True):
                result = lazy_web_element.exists(self.count)
        else:
            result = lazy_web_element.exists(self.count)

        return result

    def describe_to(self, description):
        description.append_text(
            self._create_message('Web element exists', count=self.count),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text('was not found')


element_exists = _ElementExists


class _ElementNotExists(_BasePagiumMatcher):

    def __init__(self, count: int = 1, **kwargs):
        super(_ElementNotExists, self).__init__(**kwargs)

        self.count = count

    def __matches__(self, lazy_web_element: LazyWebElement):
        driver = utils.get_driver(lazy_web_element.parent)

        if hasattr(driver, 'disable_polling'):
            with driver.disable_polling(force=True):
                result = not lazy_web_element.exists(self.count)
        else:
            result = not lazy_web_element.exists(self.count)

        return result

    def describe_to(self, description):
        description.append_text(
            self._create_message('Web element not exists', count=self.count),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text('was found')


element_not_exists = _ElementNotExists


class _URLPathEqual(_BasePagiumMatcher):

    def __init__(self, url_path: str, **kwargs):
        super(_URLPathEqual, self).__init__(**kwargs)

        self.url_path = url_path
        self.current_path = None

    def __matches__(self, browser):
        self.current_path = urlparse(browser.current_url).path
        return self.url_path == self.current_path

    def describe_to(self, description):
        description.append_text(
            self._create_message(f'Current path equal to "{self.url_path}"'),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text(f'was "{self.current_path}"')


url_path_equal = _URLPathEqual


class _URLPathContains(_BasePagiumMatcher):

    def __init__(self, url_path_part: str, **kwargs):
        super(_URLPathContains, self).__init__(**kwargs)

        self.url_path_part = url_path_part
        self.current_path = None

    def __matches__(self, browser):
        self.current_path = urlparse(browser.current_url).path
        return self.url_path_part in self.current_path

    def describe_to(self, description):
        description.append_text(
            self._create_message(f'Current path contains "{self.url_path_part}"'),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text(f'was "{self.current_path}"')


url_path_contains = _URLPathContains


class _MatchRegexp(_BasePagiumMatcher):

    def __init__(self, regexp, **kwargs):
        super(_MatchRegexp, self).__init__(**kwargs)

        self.pattern = re.compile(regexp)

    def __matches__(self, instance):
        self.text = instance.text
        return self.pattern.search(self.text) is not None

    def describe_to(self, description):
        description.append_text(
            self._create_message(f'a element text matching "{self.pattern.pattern}"'),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text(f'was "{self.text}"')


match_regexp = _MatchRegexp


class _TagAttributeEqual(_BasePagiumMatcher):

    def __init__(self, attribute_name, value, **kwargs):
        super(_TagAttributeEqual, self).__init__(**kwargs)
        self.attribute_name = attribute_name
        self.value = value

    def __matches__(self, we: Union[WebElement, LazyWebElement]):
        return we.get_attribute(self.attribute_name) == self.value

    def describe_to(self, description):
        description.append_text(
            self._create_message(f'Attribute value:"{self.value}" equals'),
        )

    def describe_mismatch(self, item, mismatch_description):
        mismatch_description.append_text('was not equals')


tag_attribute_equal = _TagAttributeEqual



# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages


setup(
    name='pagium',
    version_format='{tag}',
    setup_requires=['setuptools-git-version'],
    url='https://github.com/pagium/pagium',
    packages=find_packages(include=('pagium',)),
    description='Selenium page object implementation',
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'selenium',
        'pyhamcrest',
    ],
    classifiers=(
        'Development Status :: 4 - Beta',
        'Natural Language :: Russian',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
    ),
)
